import os
import re
import subprocess

def available_cpu_count():
    """ Number of available virtual or physical CPUs on this system, i.e.
    user/real as output by time(1) when called with an optimally scaling
    userspace-only program"""

    # cpuset
    # cpuset may restrict the number of *available* processors
    try:
        m = re.search(r'(?m)^Cpus_allowed:\s*(.*)$',
                      open('/proc/self/status').read())
        if m:
            res = bin(int(m.group(1).replace(',', ''), 16)).count('1')
            if res > 0:
                return res
    except IOError:
        pass

    # Python 2.6+
    try:
        import multiprocessing
        return multiprocessing.cpu_count()
    except (ImportError, NotImplementedError):
        pass

    # https://github.com/giampaolo/psutil
    try:
        import psutil
        return psutil.cpu_count()   # psutil.NUM_CPUS on old versions
    except (ImportError, AttributeError):
        pass

    # POSIX



import sys
import commands
import subprocess
import socket
import os

from os import path
from cpbox.app.appconfig import appconfig
from cpbox.tool import utils
from cpbox.tool import file
from cpbox.tool import logger

class DevOpsApp:
    class Error(Exception):
        """
        DevOpsApp Error
        """
        pass

    def __init__(self, app_name, log_level='info'):
        appconfig.init(app_name)
        self.env = os.environ['PUPPY_ENV'] if 'PUPPY_ENV' in os.environ else 'dev'
        syslog_ng_server = os.environ['PUPPY_SYSLOG_NG_SERVER'] if 'PUPPY_SYSLOG_NG_SERVER' in os.environ else None
        logger.make_logger(app_name, log_level, syslog_ng_server, True)
        self.logger = logger.getLogger(app_name)

        self.hostname = socket.gethostname()
        self.group_names = os.environ['PUPPY_GROUPS'].split(',') if 'PUPPY_GROUPS' in os.environ else []

        self.root_dir = path.dirname(path.realpath(sys.argv[0]))
        self.roles_dir = self.root_dir + '/roles'
        app_root_dir = self.roles_dir + '/' + app_name

        self.app_config_dir =  app_root_dir + '/config'
        self.app_templates_dir =  app_root_dir + '/templates'
        self.app_scripts_dir =  app_root_dir + '/scripts'

        self.app_storage_dir = '/opt/data/' + app_name
        self.app_persitent_storage_dir = self.app_storage_dir + '/persistent'
        self.app_runtime_storage_dir = self.app_storage_dir + '/runtime'
        self.app_logs_dir = self.app_runtime_storage_dir + '/logs'

        self._ensure_dir_and_write_permision()

        self.file_lock = None

    def run_cmd_ret(self, cmd):
        return self.run_cmd(cmd)[1]

    def run_cmd(self, cmd):
        self.logger.info('run_cmd: %s', cmd)
        ret = commands.getstatusoutput(cmd)
        return ret

    def shell_run(self, cmd):
        self.logger.info('shell_run: %s', cmd)
        return subprocess.call(cmd, shell=True)

    def _ensure_dir_and_write_permision(self):
        file.ensure_dir(self.app_storage_dir)
        file.ensure_dir(self.app_persitent_storage_dir)
        file.ensure_dir(self.app_runtime_storage_dir)
        file.ensure_dir(self.app_logs_dir)

    def _check_lock(self):
        filepath = self.app_runtime_storage_dir + '/locks/' + file.compute_lock_filepath(sys.argv)
        file_lock = file.obtain_lock(filepath)
        if file_lock is None:
            pid = 0
            with open(filepath, 'r') as f:
                pid = f.read()
            self.logger.warning('lock file exists, pid: %s => %s', pid, filepath)
            sys.exit(1)
        else:
            self.file_lock = file_lock



import argparse
import inspect
import yaml
import sys
import six

from cpbox.tool import logger as cplogger
from distutils import util

import logging
logger = logging.getLogger('functocli')
support_types = ('str', 'int', 'bool', 'float')

keep_class_set = set([])
def keep_class(cls):
    keep_class_set.add(cls)
    return cls

class AppSepc(object):

    def __init__(self, app, log_level):

        stream_handler = logging.StreamHandler()
        cplogger.setup_logger_handler('functocli', stream_handler, log_level)
        logger.setLevel(stream_handler.level)
        logger.addHandler(stream_handler)

        self.app = app
        self.public_methods = get_app_public_methods(app)
        self.method_spec_map = self._get_app_method_spec()

    def _get_app_method_spec(self):
        method_spec_map = {}
        for method_name, _ in self.public_methods.items():
            method_attr = getattr(self.app, method_name)

            args, default_values = get_func_spec(method_attr)
            comments_data = get_func_comments(method_name, method_attr, args, default_values)

            method_data = {}
            method_data['args'] = args
            method_data['default_values'] = default_values
            method_data['comments_data'] = comments_data
            method_spec_map[method_name] = method_data
        return method_spec_map


def get_func_spec(method_attr):
    func_spec = inspect.getargspec(method_attr)
    default_values = None
    if func_spec.defaults:
        default_values = dict(zip(func_spec.args[-len(func_spec.defaults):], func_spec.defaults))
    args = filter(lambda x: x != 'self', func_spec.args)
    return args, default_values

def _get_func_comments(method_attr):
    comments = inspect.getdoc(method_attr)
    if comments is None:
        return None
    else:
        return yaml.load(comments)

def get_func_comments(method_name, method_attr, args, default_values, mute=False):
    has_args = len(args) > 0
    comments_data = _get_func_comments(method_attr)
    if comments_data is None and has_args:
        logger.warning('No comment data found for method: %s', method_name)

    if not comments_data:
        comments_data = {}
    if isinstance(comments_data, six.string_types):
        logger.warning('The comment data if not yaml format: %s', method_name)
        comments_data = {}

    if 'des' not in comments_data:
        logger.warn('Missing function descripion(`des`) in the comment of method: %s', method_name)
        comments_data['des'] = ''

    if has_args:
        parameters = comments_data.get('parameters', {})
        if 'parameters' not in comments_data:
            logger.warn('Missing `parameters` field in the comment of method: %s', method_name)
        comments_data['parameters'] = _check_parameters_spec(method_name, args, parameters, default_values)
    return comments_data

def _is_float(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def _check_parameter_item_default_value(arg_name, parameter_item, default_values):
    if default_values is None or arg_name not in default_values:
        return parameter_item

    value = default_values[arg_name]
    type = _determine_type(value)

    if type is None:
        raise Error('Can not determine type for parameter: %s, parameter_item: %s' %s (arg_name, parameter_item))

    parameter_item['type'] = type
    parameter_item['des'] = 'type: %s, default value is: %s' % (type, value)
    return parameter_item

def _determine_type(value):
    type = None
    if value is None:
        type = 'str'
    elif isinstance(value, bool):
        type = 'bool'
    elif isinstance(value, six.integer_types):
        type = 'int'
    elif isinstance(value, six.string_types):
        type = 'str'
    elif _is_float(value):
        type = 'float'
    return type

def _convert_value_for_type(value, type):
    if value is None:
        return value
    elif type == 'bool':
        if isinstance(value, six.string_types):
            try:
                return bool(util.strtobool(value))
            except:
                return False
        return bool(value)
    elif type == 'int':
        return int(value)
    elif type == 'str':
        return str(value)
    elif type == 'float':
        return float(value)
    return value

def _check_parameters_spec(method_name, args, parameters, default_values):
    for arg_name in args:

        if arg_name not in parameters:
            logger.warn('No parameter config item found for argment: %s', arg_name)
            parameter_item = {}

        parameter_item = parameters.get(arg_name, {})
        if parameter_item is None:
            logger.warn('Parameter config is `None` for argment: %s', arg_name)
            parameter_item = {}

        if isinstance(parameter_item, six.string_types):
            parameter_item = {
                    'type': 'str',
                    'des': parameter_item
                    }
        parameter_item = _check_parameter_item_default_value(arg_name, parameter_item, default_values)
        if 'type' not in parameter_item:
            parameter_item['type'] = 'str'
        if 'des' not in parameter_item:
            parameter_item['des'] = arg_name

        parameters[arg_name] = parameter_item
        _check_parameter_item(arg_name, parameter_item)

    return parameters

def _check_parameter_item(arg_name, parameter_item):
    if 'type' not in parameter_item:
        logger.warn('Missing type in parameter item: %s', parameter_item)
    if parameter_item['type'] not in support_types:
        logger.warn('The type: %s is not in support types: %s', parameter_item['type'], support_types)
    if not parameter_item['des']:
        logger.warn('Missing parameter descripion(`des`): %s', parameter_item)

def get_app_public_methods(app):
    parent_members = {}
    for base_class in app.__bases__:
        if base_class in keep_class_set:
            continue
        parent_members_list = inspect.getmembers(base_class, predicate=inspect.ismethod)
        parent_members.update(dict((item[0], item[1]) for item in parent_members_list))
    public_method_names = inspect.getmembers(app, predicate=inspect.ismethod)
    ret = filter(lambda x: x[0].startswith('_') == False and x[0] not in parent_members, public_method_names)
    ret = {item[0]: item[1] for item in ret}
    return ret

def build_argparser_for_func(method_name, method_spec):

    args = method_spec['args']
    default_values = method_spec['default_values']
    comments_data = method_spec['comments_data']

    parser = argparse.ArgumentParser(usage=comments_data['des'])
    parser.add_argument('--' + method_name, action='store_true')

    parameters = comments_data.get('parameters', {})
    for arg_name in args:
        parameter = parameters.get(arg_name)
        help = parameter.get('des', '')
        metavar = ''

        if default_values is not None and arg_name in default_values:
            if len(arg_name) == 1:
                parser.add_argument('-' + arg_name, metavar=metavar, help=help, default=default_values[arg_name], nargs=1, required=False)
            else:
                parser.add_argument('--' + arg_name, metavar=metavar, help=help, default=default_values[arg_name], nargs=1, required=False)
        else:
            if len(arg_name) == 1:
                parser.add_argument('-' + arg_name, metavar=metavar, help=help, nargs=1, required=True)
            else:
                parser.add_argument('--' + arg_name, metavar=metavar, help=help, nargs=1, required=True)
    args_input, _ = parser.parse_known_args()
    args_kwargs = {}
    for arg_name in args:
        parameter = parameters.get(arg_name)
        attr = getattr(args_input, arg_name)
        if default_values is not None and arg_name in default_values and attr == default_values[arg_name]:
            args_kwargs[arg_name] = attr
        else:
            args_kwargs[arg_name] = _convert_value_for_type(
                    _get_first_from_list_or_item_self(attr), parameter['type'])
    return args_kwargs

def _get_first_from_list_or_item_self(attr):
    try:
        return next(iter(attr), None)
    except Typeerror:
        return attr

def _check_args_has(args, option):
    ret = getattr(args, option)
    return ret

def run_app(app, log_level='critical'):
    app_spec = AppSepc(app, log_level)
    parser = argparse.ArgumentParser(add_help=False)
    parser.allow_abbrev = False
    parser.add_argument('-h', '--help', action='store_true', help='show this help message and exit')
    for method_name, _ in app_spec.public_methods.items():
        parser.add_argument('--' + method_name, action='store_true', help=app_spec.method_spec_map[method_name]['comments_data']['des'])

    method_will_call = None
    if len(sys.argv) > 1:
        method_will_call = sys.argv[1].replace('--', '')
    if not method_will_call or method_will_call not in app_spec.public_methods:
        parser.print_help()
        sys.exit(0)

    args_kwargs = build_argparser_for_func(method_will_call, app_spec.method_spec_map[method_will_call])
    _app = app()
    func = getattr(_app, method_will_call)
    func(**args_kwargs)



import socket
import fcntl
import struct

def get_ip_address(ifname):
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return socket.inet_ntoa(fcntl.ioctl(
            s.fileno(),
            0x8915,  # SIOCGIFADDR
            struct.pack('256s', ifname[:15])
        )[20:24])
    except IOError:
        pass
    return None

def get_hostname_ip():
    ip = socket.gethostbyname(socket.gethostname())
    return ip



import logging
import logging.handlers
import syslog
import socket
import os
try:
    from threading import get_ident
except ImportError:
    from thread import get_ident

class SysLogHandler(logging.handlers.SysLogHandler):

    def emit(self, record):
        super(SysLogHandler, self).emit(record)

class ContextFilter(logging.Filter):

    def __init__(self):
        self.local_ip = socket.gethostbyname(socket.gethostname())

    def filter(self, record):
        record.local_ip = self.local_ip
        record.pid = os.getpid()
        record.tid = get_ident()
        return True

def getLogger(name=''):
    logger = logging.getLogger()
    return logger

def make_event_logger(app_name, syslog_ng_server):
    event_log_formatter = logging.Formatter(app_name + '_event_log: %(message)s')
    format_str = app_name + '_event_log[%(pid)d]: %(asctime)s %(local_ip)s %(pid)d.%(tid)d %(message)s'
    event_log_formatter = logging.Formatter(format_str)

    syslog_handler = None
    try:
        syslog_handler = SysLogHandler(address=(syslog_ng_server, 601), facility='local7', socktype=socket.SOCK_STREAM)
    except:
        pass
    finally:
        syslog_handler = SysLogHandler(address=(syslog_ng_server, 514), facility='local6')

    if syslog_handler is None:
        return

    filter = ContextFilter()
    syslog_handler.setFormatter(event_log_formatter)
    syslog_handler.addFilter(filter)
    event_logger = logging.getLogger('event-log')
    event_logger.setLevel(logging.INFO)
    event_logger.addHandler(syslog_handler)

def setup_logger_handler(app_name, handler, log_level_str):
    filter = ContextFilter()
    log_level = logging.INFO
    if log_level_str is not None:
        level = log_level_str.lower()
        log_level_conf = {
            'verbose': logging.NOTSET,
            'debug': logging.DEBUG,
            'info': logging.INFO,
            'warning': logging.WARNING,
            'warn': logging.WARNING,
            'error': logging.ERROR,
            'critical': logging.CRITICAL,
        }
        if level in log_level_conf:
            log_level = log_level_conf[level]

    format_str = app_name + '[%(pid)d]: %(asctime)s %(local_ip)s %(pid)d.%(tid)d %(levelname)s %(filename)s:%(lineno)d %(message)s'
    formatter = logging.Formatter(format_str)

    handler.setFormatter(formatter)
    handler.setLevel(log_level)
    handler.addFilter(filter)
    return handler

def make_logger(app_name, log_level_str=None, syslog_ng_server=None, console_debug_log=False):
    root_logger = logging.getLogger()

    basic_handlers = []
    if syslog_ng_server is not None:
        udp_syslog_handler = SysLogHandler(address=(syslog_ng_server, 514), facility='local6')
        setup_logger_handler(app_name, udp_syslog_handler, log_level_str)
        basic_handlers.append(udp_syslog_handler)
        make_event_logger(app_name, syslog_ng_server)

    if console_debug_log:
        stream_handler = logging.StreamHandler()
        setup_logger_handler(app_name, stream_handler, log_level_str)
        basic_handlers.append(stream_handler)

    root_logger.handlers = []
    for h in basic_handlers:
        root_logger.setLevel(h.level)
        root_logger.addHandler(h)



import unittest
from cpbox.tool.datatypes import NameIdDict

class TestNameIdDict(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_all(self):
        id_1 = "r-bp194452436e8b94"
        name_1 = "yizhoucp"
        item_1 = {
                "architecture_type": "cluster",
                "bandwidth": 384,
                "capacity": 16384,
                "charge_type": "PrePaid",
                "connection_domain": "r-bp194452436e8b94.redis.rds.aliyuncs.com",
                "connections": 80000,
                "create_time": "2017-08-21T09:22:01Z",
                "end_time": "2018-10-22T16:00:00Z",
                "engine_version": "2.8",
                "has_renew_change_order": False,
                "instance_class": "redis.sharding.basic.small.default",
                "instance_id": id_1,
                "instance_name": name_1
                }

        id_2 = "r-bp1e65f4b38ae304"
        name_2 = "php-session"
        item_2 = {
                "architecture_type": "standard",
                "bandwidth": 10,
                "capacity": 1024,
                "charge_type": "PrePaid",
                "connection_domain": "php-session.redis.rds.aliyuncs.com",
                "connections": 10000,
                "create_time": "2018-09-10T09:41:01Z",
                "end_time": "2018-11-10T16:00:00Z",
                "engine_version": "4.0",
                "has_renew_change_order": False,
                "instance_class": "redis.master.small.default",
                "instance_id": id_2,
                "instance_name": name_2
                }

        data = {
                id_1: item_1,
                id_2: item_2,
                }

        name_id_dict = NameIdDict('instance_id', 'instance_name', data)
        self.assertEqual(name_id_dict[id_1], item_1)
        self.assertEqual(name_id_dict[name_1], item_1)
        self.assertEqual(name_id_dict[id_2], item_2)
        self.assertEqual(name_id_dict[name_2], item_2)
        self.assertTrue(isinstance(name_id_dict, dict))

        name_id_dict = NameIdDict('instance_id', 'instance_name')
        name_id_dict.update(data)
        self.assertEqual(name_id_dict[id_1], item_1)
        self.assertEqual(name_id_dict[name_1], item_1)
        self.assertEqual(name_id_dict[id_2], item_2)
        self.assertEqual(name_id_dict[name_2], item_2)
        self.assertTrue(isinstance(name_id_dict, dict))

if __name__ =='__main__':
    unittest.main()



from cpbox.tool.utils import Singleton
import os

class ConfigData():
    __metaclass__ = Singleton

    def __init__(self):
        self._env = os.environ['PUPPY_ENV'] if 'PUPPY_ENV' in os.environ else 'dev'
        self._app_name = None
        self._data = {}

    def get_env(self):
        return self._env

    def get_app_name(self):
        return self._app_name

    def init(self, app_name, env = None, config = {}):
        self._app_name = app_name
        if env is not None:
            self._env = env
        self._data = config

    def get_config(self):
        return self._data

    def get(self, key, fallback = None):
        return self._data.get(key, fallback)

appconfig = ConfigData()



import commands
import shlex
import subprocess

def docker0_ip():
    cmd="ip addr show docker0 | grep -Eo 'inet (addr:)?([0-9]*\.){3}[0-9]*' | grep -Eo '([0-9]*\.){3}[0-9]*' | grep -v '127.0.0.1' | awk '{print $1}' | head  -1"
    ip = run_cmd_ret(cmd)
    if not ip:
        ip = '172.20.0.1'
    return ip

def run_cmd(cmd, mute=False):
    if not mute:
        print cmd
    return commands.getstatusoutput(cmd)

def call(cmd, mute=False):
    if not mute:
        print cmd
    args = ['/bin/bash', '-c', cmd]
    return subprocess.call(args)

def run_cmd_ret(cmd, mute=False):
    return run_cmd(cmd, mute)[1]

def stop_container(container_name):
    cmd="docker ps -a -f name='^/%s$' | grep '%s' | awk '{print $1}' | xargs -I {} docker rm -f --volumes {}" % (container_name, container_name)
    call(cmd)

def firewall_list_public_access():
    code, list = run_cmd('firewall-cmd --zone=public --list-port')
    if code == 0:
        list = list.split()
        list.sort()
        return list
    return []

def firewall_ensure_public_access(ports):
    for port in ports:
        run_cmd("firewall-cmd --zone=public --add-port=%s --permanent" % port)
    if len(ports) > 0:
        run_cmd("firewall-cmd --reload")

def firewall_remove_public_access(ports):
    for port in ports:
        run_cmd("firewall-cmd --zone=public --remove-port=%s --permanent" % port)
    if len(ports) > 0:
        run_cmd("firewall-cmd --reload")

def run_cmd_(cmd):
    print cmd



import unittest

from cpbox.app import eventlog
from cpbox.app import appconfig
from datetime import datetime
from cpbox.tool import logger
import time

class TestToolUtils(unittest.TestCase):

    def setUp(self):
        logger.make_logger('test-event-log', 'debug', 'localhost', True)

    def test_event_log(self):
        for i in range(1, 10):
            payload = {'time': str(datetime.now())}
            eventlog.add_event_log('test', payload)
        eventlog.log_worker.wait()

if __name__ == '__main__':
    unittest.main()
